from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.response import Response

from crosswalk.authentication import AuthenticatedView
from crosswalk.models import Domain, Entity
from crosswalk.serializers import EntitySerializer
from crosswalk.utils import import_class


class AliasOrCreate(AuthenticatedView):

    def post(self, request, domain):
        """
        Create an alias if an entity is found above a certain match threshold
        or create a new entity.
        """
        user = request.user
        data = request.data.copy()
        query_field = data.get('query_field')
        query_value = data.get('query_value')
        block_attrs = data.get('block_attrs', {})
        create_attrs = data.get('create_attrs', {})
        return_canonical = data.get('return_canonical', True)
        threshold = data.get('threshold')
        scorer_class = data.get('scorer', 'fuzzywuzzy.default_process')
        scorer = import_class('crosswalk.scorers.{}'.format(scorer_class))

        try:
            domain = Domain.objects.get(slug=domain)
        except ObjectDoesNotExist:
            return Response(
                "Domain not found.",
                status=status.HTTP_404_NOT_FOUND
            )

        # Find the best match for a query
        entities = Entity.objects.filter(domain=domain)
        entities = entities.filter(attributes__contains=block_attrs)

        entity_values = [e.attributes[query_field] for e in entities]
        match, score = scorer(query_value, entity_values)

        entities = entities.filter(
            **{'attributes__{}'.format(query_field): match}
        )

        if entities.count() > 1:
            return Response(
                "More than one alias candiate for entity.",
                status=status.HTTP_403_FORBIDDEN
            )

        entity = entities.first()

        attributes = {
            **{query_field: query_value},
            **block_attrs,
            **create_attrs
        }

        if entity.attributes == attributes:
            return Response(
                "Entity appears to already exist.",
                status=status.HTTP_409_CONFLICT
            )

        if score > threshold:
            aliased = True
            alias = Entity(
                attributes=attributes,
                alias_for=entity,
                created_by=user,
                domain=domain
            )
            alias.save()
            if return_canonical:
                while entity.alias_for:
                    entity = entity.alias_for
        else:
            aliased = False
            entity = Entity(
                attributes=attributes,
                created_by=user,
                domain=domain
            )
            entity.save()

        return Response({
            "entity": EntitySerializer(entity).data,
            "created": True,
            "aliased": aliased,
            "match_score": score,
        }, status=status.HTTP_200_OK)



from django.contrib.auth.models import User
from django.db import models
from uuslug import uuslug


class Domain(models.Model):
    slug = models.SlugField(
        blank=True, max_length=250, unique=True, editable=False)

    name = models.CharField(max_length=250, unique=True)

    parent = models.ForeignKey(
        'self',
        null=True,
        blank=True,
        related_name='children',
        on_delete=models.PROTECT
    )

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    created_by = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        on_delete=models.SET_NULL
    )

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = uuslug(
                self.name,
                instance=self,
                max_length=250,
                separator='-',
                start_no=2
            )
        super(Domain, self).save(*args, **kwargs)

    def __str__(self):
        return self.name



from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.response import Response

from crosswalk.authentication import AuthenticatedView
from crosswalk.models import Domain, Entity
from crosswalk.serializers import EntitySerializer
from crosswalk.utils import import_class


class BestMatchOrCreate(AuthenticatedView):
    def post(self, request, domain):
        """
        Get the best matched entity for a given query or create an entity if
        one is not found above a certain match threshold.

        If the entity is an alias of another entity, the aliased entity is
        returned.
        """
        user = request.user
        data = request.data.copy()
        query_field = data.get('query_field')
        query_value = data.get('query_value')
        block_attrs = data.get('block_attrs', {})
        create_attrs = data.get('create_attrs', {})
        threshold = data.get('threshold')
        return_canonical = data.get('return_canonical', True)
        scorer_class = data.get('scorer', 'fuzzywuzzy.default_process')
        scorer = import_class('crosswalk.scorers.{}'.format(scorer_class))

        try:
            domain = Domain.objects.get(slug=domain)
        except ObjectDoesNotExist:
            return Response(
                "Domain not found.",
                status=status.HTTP_404_NOT_FOUND
            )

        entities = Entity.objects.filter(domain=domain)
        entities = entities.filter(attributes__contains=block_attrs)

        entity_values = [e.attributes[query_field] for e in entities]
        match, score = scorer(query_value, entity_values)

        entity = entities.filter(
            **{'attributes__{}'.format(query_field): match}
        ).first()

        created = False
        aliased = False

        if score < threshold:
            created = True
            uuid = create_attrs.pop('uuid', None)
            entity = Entity(
                uuid=uuid,
                attributes={
                    **{query_field: query_value},
                    **block_attrs,
                    **create_attrs,
                },
                created_by=user,
                domain=domain
            )
            entity.save()

        if return_canonical:
            while entity.alias_for:
                aliased = True
                entity = entity.alias_for

        return Response({
            "entity": EntitySerializer(entity).data,
            "created": created,
            "match_score": score,
            "aliased": aliased,
        }, status=status.HTTP_200_OK)



from crosswalk.models import ApiUser, Domain, Entity
from django.contrib import admin


class ApiUserAdmin(admin.ModelAdmin):
    readonly_fields = ('token',)
    fields = ('user', 'token',)


admin.site.register(ApiUser, ApiUserAdmin)
admin.site.register(Domain)
admin.site.register(Entity)



from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import serializers

from .models import Domain, Entity


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = (
            'username',
            'first_name',
            'last_name',
            'email',
        )


class DomainSerializer(serializers.ModelSerializer):
    name = serializers.CharField(validators=[])
    parent = serializers.SlugRelatedField(
        slug_field='slug',
        queryset=Domain.objects.all(),
        required=False,
        allow_null=True,
    )

    def create(self, validated_data):
        try:
            parent = Domain.objects.get(
                slug=validated_data.get('parent')
            )
        except ObjectDoesNotExist:
            parent = None
        domain, created = Domain.objects.get_or_create(
            name=validated_data.get('name'),
            defaults={"parent": parent}
        )
        return domain

    class Meta:
        model = Domain
        lookup_field = 'slug'
        fields = (
            'slug',
            'name',
            'parent',
        )


class EntityListSerializer(serializers.ListSerializer):
    def create(self, validated_data):
        entities = [Entity(**data) for data in validated_data]
        return Entity.objects.bulk_create(entities)


class EntitySerializer(serializers.ModelSerializer):
    uuid = serializers.UUIDField(
        format='hex_verbose',
        required=False,
    )
    domain = serializers.SlugRelatedField(
        slug_field='slug',
        queryset=Domain.objects.all(),
    )
    alias_for = serializers.PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        required=False,
        allow_null=True,
        pk_field=serializers.UUIDField(format='hex_verbose'),
    )
    superseded_by = serializers.PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        required=False,
        allow_null=True,
        pk_field=serializers.UUIDField(format='hex_verbose'),
    )

    class Meta:
        model = Entity
        list_serializer_class = EntityListSerializer
        fields = (
            'uuid',
            'attributes',
            'domain',
            'superseded_by',
            'alias_for',
        )



from django.core.exceptions import ObjectDoesNotExist
from rest_framework import authentication, exceptions
from rest_framework.permissions import IsAuthenticated
from rest_framework.views import APIView

from .models import ApiUser


class TokenAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        # Of format "TOKEN <SOME TOKEN>"
        token = request.META.get('HTTP_AUTHORIZATION', None)
        if not token:
            raise exceptions.AuthenticationFailed(
                'Missing authorization header'
            )
        try:
            user = ApiUser.objects.get(token=token.split(' ')[1]).user
        except ObjectDoesNotExist:
            raise exceptions.AuthenticationFailed('Unauthorized')
        return (user, None)


class AuthenticatedView(APIView):
    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)
